# Author: Toshio Kuratomi <tkuratom@redhat.com>
# GNU General Public License v3.0+ (see LICENSES/GPL-3.0-or-later.txt or
# https://www.gnu.org/licenses/gpl-3.0.txt)
# SPDX-License-Identifier: GPL-3.0-or-later
# SPDX-FileCopyrightText: 2020, Ansible Project
"""Entrypoint to the antsibull-docs script."""

import asyncio
import os
import os.path
import tempfile
import textwrap
import typing as t
from collections import defaultdict
from concurrent.futures import ProcessPoolExecutor

import aiohttp
import asyncio_pool  # type: ignore[import]
from antsibull_core.ansible_core import get_ansible_core
from antsibull_core.collections import install_together
from antsibull_core.compat import asyncio_run, best_get_loop
from antsibull_core.dependency_files import DepsFile
from antsibull_core.galaxy import CollectionDownloader
from antsibull_core.logging import log
from antsibull_core.venv import FakeVenvRunner, VenvRunner
from pydantic import ValidationError

from ... import app_context
from ...augment_docs import augment_docs
from ...collection_links import load_collections_links
from ...docs_parsing.fqcn import get_fqcn_parts
from ...docs_parsing.parsing import get_ansible_plugin_info
from ...docs_parsing.routing import (
    find_stubs,
    load_all_collection_routing,
    remove_redirect_duplicates,
)
from ...env_variables import collect_referenced_environment_variables, load_ansible_config
from ...extra_docs import load_collections_extra_docs
from ...schemas.docs import DOCS_SCHEMAS
from ...utils.collection_name_transformer import CollectionNameTransformer
from ...write_docs.collections import output_extra_docs, output_indexes
from ...write_docs.hierarchy import output_collection_index, output_collection_namespace_indexes
from ...write_docs.indexes import (
    output_callback_indexes,
    output_environment_variables,
    output_plugin_indexes,
)
from ...write_docs.plugin_stubs import output_all_plugin_stub_rst
from ...write_docs.plugins import output_all_plugin_rst

if t.TYPE_CHECKING:
    import semantic_version as semver  # pylint:disable=unused-import


mlog = log.fields(mod=__name__)

#: Mapping of plugins to nonfatal errors.  This is the type to use when returning the mapping.
PluginErrorsRT = t.DefaultDict[str, t.DefaultDict[str, t.List[str]]]


def get_collection_namespaces(collection_names: t.Iterable[str]) -> t.Dict[str, t.List[str]]:
    """
    Return the plugins which are in each collection.

    :arg collection_names: An iterable of collection names.
    :returns: Mapping from collection namespaces to list of collection names.
    """
    namespaces = defaultdict(list)
    for collection_name in collection_names:
        namespace, name = collection_name.split('.', 1)
        namespaces[namespace].append(name)
    return namespaces


async def retrieve(ansible_core_version: str,
                   collections: t.Mapping[str, str],
                   tmp_dir: str,
                   galaxy_server: str,
                   ansible_core_source: t.Optional[str] = None,
                   collection_cache: t.Optional[str] = None) -> t.Dict[str, 'semver.Version']:
    """
    Download ansible-core and the collections.

    :arg ansible_core_version: Version of ansible-core to download.
    :arg collections: Map of collection names to collection versions to download.
    :arg tmp_dir: The directory to download into.
    :arg galaxy_server: URL to the galaxy server.
    :kwarg ansible_core_source: If given, a path to an ansible-core checkout or expanded sdist.
        This will be used instead of downloading an ansible-core package if the version matches
        with ``ansible_core_version``.
    :kwarg collection_cache: If given, a path to a directory containing collection tarballs.
        These tarballs will be used instead of downloading new tarballs provided that the
        versions match the criteria (latest compatible version known to galaxy).
    :returns: Map of collection name to directory it is in.  ansible-core will
        use the special key, `_ansible_core`.
    """
    collection_dir = os.path.join(tmp_dir, 'collections')
    os.mkdir(collection_dir, mode=0o700)

    requestors = {}

    lib_ctx = app_context.lib_ctx.get()
    async with aiohttp.ClientSession() as aio_session:
        async with asyncio_pool.AioPool(size=lib_ctx.thread_max) as pool:
            requestors['_ansible_core'] = await pool.spawn(
                get_ansible_core(aio_session, ansible_core_version, tmp_dir,
                                 ansible_core_source=ansible_core_source))

            downloader = CollectionDownloader(aio_session, collection_dir,
                                              galaxy_server=galaxy_server,
                                              collection_cache=collection_cache)
            for collection, version in collections.items():
                requestors[collection] = await pool.spawn(
                    downloader.download(collection, version))

            responses = await asyncio.gather(*requestors.values())

    # Note: Python dicts have always had a stable order as long as you don't modify the dict.
    # So requestors (implicitly, the keys) and responses have a matching order here.
    return dict(zip(requestors, responses))


def normalize_plugin_info(plugin_type: str,
                          plugin_info: t.Mapping[str, t.Any]
                          ) -> t.Tuple[t.Dict[str, t.Any], t.List[str]]:
    """
    Normalize and validate all of the plugin docs.

    :arg plugin_type: The type of plugins that we're getting docs for.
    :arg plugin_info: Mapping of plugin_info.  The toplevel keys are plugin names.
        See the schema in :mod:`antsibull.schemas.docs` for what the data should look like and just
        how much conversion we can perform on it.
    :returns: A tuple containing a "copy" of plugin_info with all of the data normalized and a list
        of nonfatal errors.  The plugin_info dict will follow the structure expressed in the schemas
        in :mod:`antsibull.schemas.docs`.  The nonfatal errors are strings representing the problems
        encountered.
    """
    # If you wonder why this code isn't showing up in code coverage: that's because it's executed
    # in a subprocess. See normalize_all_plugin_info below.

    if 'error' in plugin_info:
        return ({}, [plugin_info['error']])

    errors: t.List[str] = []
    if plugin_type == 'role':
        try:
            parsed = DOCS_SCHEMAS[plugin_type].parse_obj(plugin_info)  # type: ignore[attr-defined]
            return parsed.dict(by_alias=True), errors
        except ValidationError as e:
            raise ValueError(str(e))  # pylint:disable=raise-missing-from

    new_info: t.Dict[str, t.Any] = {}
    # Note: loop through "doc" before any other keys.
    for field in ('doc', 'examples', 'return'):
        try:
            schema = DOCS_SCHEMAS[plugin_type][field]  # type: ignore[index]
            field_model = schema.parse_obj({field: plugin_info.get(field)})
        except ValidationError as e:
            if field == 'doc':
                # We can't recover if there's not a doc field
                # pydantic exceptions are not picklable (probably due to bugs in the pickle module)
                # so convert it to an exception type which is picklable
                raise ValueError(str(e))  # pylint:disable=raise-missing-from

            # But we can use the default value (some variant of "empty") for everything else
            # Note: We looped through doc first and returned an exception if doc did not normalize
            # so we're able to use it in the error message here.
            errors.append(f'Unable to normalize {new_info["doc"]["name"]}: {field}'
                          f' due to: {str(e)}')

            field_model = DOCS_SCHEMAS[plugin_type][field].parse_obj({})  # type: ignore[index]

        new_info.update(field_model.dict(by_alias=True))

    return (new_info, errors)


async def normalize_all_plugin_info(plugin_info: t.Mapping[str, t.Mapping[str, t.Any]]
                                    ) -> t.Tuple[t.Dict[str, t.MutableMapping[str, t.Any]],
                                                 PluginErrorsRT]:
    """
    Normalize the data in plugin_info so that it is ready to be passed to the templates.

    :arg plugin_info: Mapping of information about plugins.  This contains information about all of
        the plugins that are to be documented. See the schema in :mod:`antsibull.schemas.docs` for
        the structure of the information.
    :returns: A tuple of plugin_info (this is a "copy" of the input plugin_info with all of the
        data normalized) and a mapping of errors.  The plugin_info may have less records than the
        input plugin_info if there were plugin records which failed to validate.  The mapping of
        errors takes the form of:

        .. code-block:: yaml

            plugin_type:
                plugin_name:
                    - error string
                    - error string
    """
    loop = best_get_loop()
    lib_ctx = app_context.lib_ctx.get()
    executor = ProcessPoolExecutor(max_workers=lib_ctx.process_max)

    # Normalize each plugin in a subprocess since normalization is CPU bound
    normalizers = {}
    for plugin_type, plugin_list_for_type in plugin_info.items():
        for plugin_name, plugin_record in plugin_list_for_type.items():
            normalizers[(plugin_type, plugin_name)] = loop.run_in_executor(
                executor, normalize_plugin_info, plugin_type, plugin_record)

    results = await asyncio.gather(*normalizers.values(), return_exceptions=True)

    new_plugin_info: t.DefaultDict[str, t.MutableMapping[str, t.Any]]
    new_plugin_info = defaultdict(dict)  # pyre-ignore[9]
    nonfatal_errors: PluginErrorsRT = defaultdict(lambda: defaultdict(list))
    for (plugin_type, plugin_name), plugin_record in zip(normalizers, results):
        # Errors which broke doc parsing (and therefore we won't have enough info to
        # build a docs page)
        if isinstance(plugin_record, Exception):
            # An exception means there is no usable documentation for this plugin
            # Record a nonfatal error and then move on
            nonfatal_errors[plugin_type][plugin_name].append(str(plugin_record))
            continue

        # Errors where we have at least docs.  We can still create a docs page for these with some
        # information left out
        if plugin_record[1]:
            nonfatal_errors[plugin_type][plugin_name].extend(plugin_record[1])

        new_plugin_info[plugin_type][plugin_name] = plugin_record[0]

    return new_plugin_info, nonfatal_errors


def get_plugin_contents(plugin_info: t.Mapping[str, t.Mapping[str, t.Any]],
                        nonfatal_errors: PluginErrorsRT
                        ) -> t.DefaultDict[str, t.DefaultDict[str, t.Dict[str, str]]]:
    """
    Return the collections with their plugins for every plugin type.

    :arg plugin_info: Mapping of plugin type to a mapping of plugin name to plugin record.
        The plugin_type, plugin_name, and short_description from plugin_records are used.
    :arg nonfatal_errors: mapping of plugin type to plugin name to list of error messages.
        The plugin_type and plugin_name are used.
    :returns: A Mapping of plugin type to a mapping of collection name to a mapping of plugin names
        to short_descriptions.
    plugin_type:
        collection:
            - plugin_short_name: short_description
    """
    plugin_contents: t.DefaultDict[str, t.DefaultDict[str, t.Dict[str, str]]]
    plugin_contents = defaultdict(lambda: defaultdict(dict))
    # Some plugins won't have an entry in the plugin_info because documentation failed to parse.
    # Those should be documented in the nonfatal_errors information.
    for plugin_type, plugin_list in nonfatal_errors.items():
        for plugin_name, dummy_ in plugin_list.items():
            namespace, collection, short_name = get_fqcn_parts(plugin_name)
            plugin_contents[plugin_type]['.'.join((namespace, collection))][short_name] = ''

    for plugin_type, plugin_dict in plugin_info.items():
        for plugin_name, plugin_desc in plugin_dict.items():
            namespace, collection, short_name = get_fqcn_parts(plugin_name)
            if plugin_type == 'role':
                desc = ''
                if 'entry_points' in plugin_desc and 'main' in plugin_desc['entry_points']:
                    desc = plugin_desc['entry_points']['main'].get('short_description') or ''
            elif 'doc' in plugin_desc:
                desc = plugin_desc['doc'].get('short_description') or ''
            else:
                desc = ''
            plugin_contents[plugin_type]['.'.join((namespace, collection))][short_name] = desc

    return plugin_contents


def get_callback_plugin_contents(plugin_info: t.Mapping[str, t.Mapping[str, t.Any]],
                                 ) -> t.DefaultDict[str, t.DefaultDict[str, t.Dict[str, str]]]:
    """
    Return the collections with their plugins for every callback plugin type.

    :arg plugin_info: Mapping of plugin type to a mapping of plugin name to plugin record.
        The plugin_type, plugin_name, and short_description from plugin_records are used.
    :returns: A Mapping of callback plugin type to a mapping of collection name to a mapping of
        plugin names to short_descriptions.
    callback_type:
        collection:
            - plugin_short_name: short_description
    """
    callback_plugin_contents: t.DefaultDict[str, t.DefaultDict[str, t.Dict[str, str]]]
    callback_plugin_contents = defaultdict(lambda: defaultdict(dict))

    if plugin_info.get('callback'):
        for plugin_name, plugin_desc in plugin_info['callback'].items():
            if 'doc' in plugin_desc:
                desc = plugin_desc['doc'].get('short_description') or ''
                callback_type = plugin_desc['doc'].get('type') or ''
                if callback_type:
                    namespace, collection, short_name = get_fqcn_parts(plugin_name)
                    collection_name = '.'.join((namespace, collection))
                    callback_plugin_contents[callback_type][collection_name][short_name] = desc

    return callback_plugin_contents


def get_collection_contents(plugin_content: t.Mapping[str, t.Mapping[str, t.Mapping[str, str]]],
                            ) -> t.DefaultDict[str, t.Dict[str, t.Mapping[str, str]]]:
    """
    Return the plugins which are in each collection.

    :arg plugin_content: Mapping of plugin type to a mapping of collection name to a mapping of
        plugin name to short description.
    :returns: A Mapping of collection name to a mapping of plugin type to a mapping of plugin names
        to short_descriptions.
    collection:
        plugin_type:
            - plugin_short_name: short_description
    """
    collection_plugins: t.DefaultDict[str, t.Dict[str, t.Mapping[str, str]]] = defaultdict(dict)

    for plugin_type, collection_data in plugin_content.items():
        for collection_name, plugin_data in collection_data.items():
            collection_plugins[collection_name][plugin_type] = plugin_data

    return collection_plugins


def generate_docs_for_all_collections(venv: t.Union[VenvRunner, FakeVenvRunner],
                                      collection_dir: t.Optional[str],
                                      dest_dir: str,
                                      collection_names: t.Optional[t.List[str]] = None,
                                      create_indexes: bool = True,
                                      squash_hierarchy: bool = False,
                                      breadcrumbs: bool = True,
                                      use_html_blobs: bool = False,
                                      fail_on_error: bool = False,
                                      for_official_docsite: bool = False) -> int:
    """
    Create documentation for a set of installed collections.

    :arg venv: The venv in which ansible-core is installed.
    :arg collection_dir: The directory in which the collections have been installed.
                         If ``None``, the collections are assumed to be in the current
                         search path for Ansible.
    :arg dest_dir: The directory into which the documentation is written.
    :kwarg collection_names: Optional list of collection names. If specified, only documentation
                             for these collections will be collected and generated.
    :kwarg create_indexes: Whether to create the collection, namespace, and plugin indexes. By
                           default, they are created.
    :kwarg squash_hierarchy: If set to ``True``, no directory hierarchy will be used.
                             Undefined behavior if documentation for multiple collections are
                             created.
    :kwarg breadcrumbs: Default True.  Set to False if breadcrumbs for collections should be
        disabled.  This will disable breadcrumbs but save on memory usage.
    :kwarg use_html_blobs: Default False.  Set to True if HTML blobs should be used instead of
        RST tables for parameter and return value tables.
    :kwarg fail_on_error: Default False.  Set to True to fail on loading or schema validation
        errors, instead of generating error pages.
    :kwarg for_official_docsite: Default False.  Set to True to use wording specific for the
        official docsite on docs.ansible.com.
    :returns: A return code for the program.  See :func:`antsibull.cli.antsibull_docs.main` for
        details on what each code means.
    """
    flog = mlog.fields(func='generate_docs_for_all_collections')
    flog.notice('Begin')

    app_ctx = app_context.app_ctx.get()

    # Get the info from the plugins
    plugin_info, full_collection_metadata = asyncio_run(get_ansible_plugin_info(
        venv, collection_dir, collection_names=collection_names))
    flog.notice('Finished parsing info from plugins and collections')
    # flog.fields(plugin_info=plugin_info).debug('Plugin data')
    # flog.fields(
    #     collection_metadata=full_collection_metadata).debug('Collection metadata')

    collection_metadata = dict(full_collection_metadata)
    if collection_names is not None and 'ansible.builtin' not in collection_names:
        del collection_metadata['ansible.builtin']

    # Load collection routing information
    collection_routing = asyncio_run(load_all_collection_routing(collection_metadata))
    flog.notice('Finished loading collection routing information')
    # flog.fields(collection_routing=collection_routing).debug('Collection routing infos')

    remove_redirect_duplicates(plugin_info, collection_routing)
    stubs_info = find_stubs(plugin_info, collection_routing)
    # flog.fields(stubs_info=stubs_info).debug('Stubs info')

    new_plugin_info, nonfatal_errors = asyncio_run(normalize_all_plugin_info(plugin_info))
    flog.fields(errors=len(nonfatal_errors)).notice('Finished data validation')
    augment_docs(new_plugin_info)
    flog.notice('Finished calculating new data')

    # Load collection extra docs data
    extra_docs_data = asyncio_run(load_collections_extra_docs(
        {name: data.path for name, data in collection_metadata.items()}))
    flog.debug('Finished getting collection extra docs data')

    # Load collection links data
    link_data = asyncio_run(load_collections_links(
        {name: data.path for name, data in collection_metadata.items()}))
    flog.debug('Finished getting collection link data')

    plugin_contents = get_plugin_contents(new_plugin_info, nonfatal_errors)
    callback_plugin_contents = get_callback_plugin_contents(new_plugin_info)
    collection_to_plugin_info = get_collection_contents(plugin_contents)
    # Make sure collections without documentable plugins are mentioned
    for collection in collection_metadata:
        collection_to_plugin_info[collection]  # pylint:disable=pointless-statement
    flog.debug('Finished getting collection data')

    # Fail on errors
    if fail_on_error and nonfatal_errors:
        print("Found errors in some modules or plugins:")
        for plugin_type, plugins in sorted(nonfatal_errors.items()):
            for plugin_name, errors in sorted(plugins.items()):
                for error in errors:
                    print(f"{plugin_name} {plugin_type}: {textwrap.indent(error, '    ').lstrip()}")
        return 1

    # Handle environment variables
    ansible_config = load_ansible_config(full_collection_metadata['ansible.builtin'])
    referenced_env_vars = collect_referenced_environment_variables(new_plugin_info, ansible_config)

    collection_namespaces = get_collection_namespaces(collection_to_plugin_info.keys())

    collection_url = CollectionNameTransformer(
        app_ctx.collection_url, 'https://galaxy.ansible.com/{namespace}/{name}')
    collection_install = CollectionNameTransformer(
        app_ctx.collection_install, 'ansible-galaxy collection install {namespace}.{name}')

    # Only build top-level index if requested
    if create_indexes:
        asyncio_run(output_collection_index(collection_to_plugin_info, collection_namespaces,
                                            dest_dir, collection_url=collection_url,
                                            collection_install=collection_install,
                                            breadcrumbs=breadcrumbs,
                                            for_official_docsite=for_official_docsite))
        flog.notice('Finished writing collection index')
        asyncio_run(output_collection_namespace_indexes(collection_namespaces, dest_dir,
                                                        collection_url=collection_url,
                                                        collection_install=collection_install,
                                                        breadcrumbs=breadcrumbs,
                                                        for_official_docsite=for_official_docsite))
        flog.notice('Finished writing collection namespace index')
        asyncio_run(output_plugin_indexes(plugin_contents, dest_dir,
                                          collection_url=collection_url,
                                          collection_install=collection_install,
                                          for_official_docsite=for_official_docsite))
        flog.notice('Finished writing plugin indexes')
        asyncio_run(output_callback_indexes(callback_plugin_contents,
                                            dest_dir, collection_url=collection_url,
                                            collection_install=collection_install,
                                            for_official_docsite=for_official_docsite))
        flog.notice('Finished writing callback plugin indexes')

    asyncio_run(output_indexes(collection_to_plugin_info, dest_dir,
                               collection_url=collection_url,
                               collection_install=collection_install,
                               collection_metadata=collection_metadata,
                               squash_hierarchy=squash_hierarchy,
                               extra_docs_data=extra_docs_data,
                               link_data=link_data,
                               breadcrumbs=breadcrumbs,
                               for_official_docsite=for_official_docsite))
    flog.notice('Finished writing indexes')

    asyncio_run(output_all_plugin_stub_rst(stubs_info, dest_dir,
                                           collection_url=collection_url,
                                           collection_install=collection_install,
                                           collection_metadata=collection_metadata,
                                           link_data=link_data,
                                           squash_hierarchy=squash_hierarchy,
                                           for_official_docsite=for_official_docsite))
    flog.debug('Finished writing plugin stubs')

    asyncio_run(output_all_plugin_rst(collection_to_plugin_info, new_plugin_info,
                                      nonfatal_errors, dest_dir,
                                      collection_url=collection_url,
                                      collection_install=collection_install,
                                      collection_metadata=collection_metadata,
                                      link_data=link_data,
                                      squash_hierarchy=squash_hierarchy,
                                      use_html_blobs=use_html_blobs,
                                      for_official_docsite=for_official_docsite))
    flog.debug('Finished writing plugin docs')

    asyncio_run(output_extra_docs(dest_dir, extra_docs_data,
                                  squash_hierarchy=squash_hierarchy))
    flog.debug('Finished writing extra docs')

    asyncio_run(output_environment_variables(dest_dir, referenced_env_vars,
                                             squash_hierarchy=squash_hierarchy))
    flog.debug('Finished writing environment variables')
    return 0


def generate_docs() -> int:
    """
    Create documentation for the stable subcommand.

    Stable documentation creates documentation for a built version of Ansible.  It uses the exact
    versions of collections included in the last Ansible release to generate rst files documenting
    those collections.

    :returns: A return code for the program.  See :func:`antsibull.cli.antsibull_docs.main` for
        details on what each code means.
    """
    flog = mlog.fields(func='generate_docs')
    flog.notice('Begin generating docs')

    app_ctx = app_context.app_ctx.get()

    # Parse the deps file
    flog.fields(deps_file=app_ctx.extra['deps_file']).info('Parse deps file')
    deps_file = DepsFile(app_ctx.extra['deps_file'])
    dummy_, ansible_core_version, collections = deps_file.parse()
    collections.pop('_python', None)
    flog.debug('Finished parsing deps file')

    with tempfile.TemporaryDirectory() as tmp_dir:
        # Retrieve ansible-core and the collections
        flog.fields(tmp_dir=tmp_dir).info('created tmpdir')
        collection_tarballs = asyncio_run(
            retrieve(ansible_core_version, collections, tmp_dir,
                     galaxy_server=app_ctx.galaxy_url,
                     ansible_core_source=app_ctx.extra['ansible_core_source'],
                     collection_cache=app_ctx.collection_cache))
        # flog.fields(tarballs=collection_tarballs).debug('Download complete')
        flog.notice('Finished retrieving tarballs')

        # Get the ansible-core location
        try:
            ansible_core_path = collection_tarballs.pop('_ansible_core')
        except KeyError:
            print('ansible-core did not download successfully')
            return 3
        flog.fields(ansible_core_path=ansible_core_path).info('ansible-core location')

        # Install the collections to a directory

        # Directory that ansible needs to see
        collection_dir = os.path.join(tmp_dir, 'installed')
        # Directory that the collections will be untarred inside of
        collection_install_dir = os.path.join(collection_dir, 'ansible_collections')
        # Safe to recursively mkdir because we created the tmp_dir
        os.makedirs(collection_install_dir, mode=0o700)
        flog.fields(collection_install_dir=collection_install_dir).debug('collection install dir')

        # Install the collections
        asyncio_run(install_together(list(collection_tarballs.values()), collection_install_dir))
        flog.notice('Finished installing collections')

        # Create venv for ansible-core
        venv = VenvRunner('ansible-core-venv', tmp_dir)
        venv.install_package(ansible_core_path)
        flog.fields(venv=venv).notice('Finished installing ansible-core')

        return generate_docs_for_all_collections(
            venv, collection_dir, app_ctx.extra['dest_dir'],
            breadcrumbs=app_ctx.breadcrumbs,
            use_html_blobs=app_ctx.use_html_blobs,
            fail_on_error=app_ctx.extra['fail_on_error'],
            for_official_docsite=True)
